import logging
from copy import deepcopy
from itertools import product

import click
from zpy.files import read_json, to_pathlib_path

from cli.config import initialize_config, login, read_config, switch_env
from cli.datasets import (create_generated_dataset, create_uploaded_dataset,
                          fetch_dataset, fetch_datasets, filter_dataset)
from cli.jobs import create_new_job, fetch_jobs
from cli.sims import create_sim, fetch_sim, fetch_sims
from cli.utils import parse_args

log = logging.getLogger(__name__)


@click.group()
@click.option('--verbose', '-v', is_flag=True, help="Enables verbose mode.")
def cli(verbose=False):
    ''' zpy cli is client side ragnarok '''
    # Set up logging
    level = logging.INFO
    if verbose:
        level = logging.DEBUG
    logging.basicConfig(
        level=level, format='%(message)s')
    initialize_config()


@cli.command('help')
def _help():
    log.info('usage: zpy <command> [<args>]')


@cli.command('env')
@click.argument('env', type=click.Choice(['local', 'stage', 'prod']))
def _env(env):
    """ switch target environment """
    log.info(f'switching to environment {env}')
    switch_env(env)


@cli.command('login')
@click.argument('username')
@click.argument('password')
def _login(username, password):
    login(username, password)


@cli.command('config')
def _config():
    """ display config """
    config = read_config()
    log.info('zpy cli configuration:')
    log.info(config)


############
### LIST ###
############


@cli.group()
def list():
    """ list resources """
    pass


@list.command('datasets')
def list_datasets():
    fetch_datasets()


@list.command('sims')
def list_sims():
    fetch_sims()


@list.command('jobs')
def list_jobs():
    fetch_jobs()


###########
### GET ###
###########


@cli.group()
def get():
    """ get resource """
    pass


@get.command('dataset')
@click.argument('name')
@click.argument('dtype', type=click.Choice(['job', 'generated', 'uploaded']))
@click.argument('path')
def get_dataset(name, dtype, path):
    dir_path = to_pathlib_path(path)
    if not dir_path.exists():
        log.info(f'output path {dir_path} does not exist')
        return
    fetch_dataset(name, path, dtype)


@get.command('sim')
@click.argument('name')
@click.argument('path')
def get_sim(name, path):
    dir_path = to_pathlib_path(path)
    if not dir_path.exists():
        log.info(f'output path {dir_path} does not exist')
        return
    fetch_sim(name, path)

##############
### UPLOAD ###
##############


@cli.group()
def upload():
    """ upload resource """
    pass


@upload.command('sim')
@click.argument('name')
@click.argument('path')
def upload_sim(name, path):
    input_path = to_pathlib_path(path)
    if not input_path.exists():
        log.warning(f'input path {input_path} does not exist')
        return
    if input_path.suffix != '.zip':
        log.warning(f'input path {input_path} not a zip file')
    create_sim(name, path)


@upload.command('dataset')
@click.argument('name')
@click.argument('path')
def upload_dataset(name, path):
    input_path = to_pathlib_path(path)
    if not input_path.exists():
        log.info(f'input path {input_path} does not exist')
        return
    if input_path.suffix != '.zip':
        log.warning(f'input path {input_path} not a zip file')
    create_uploaded_dataset(name, path)


##############
### CREATE ###
##############


@cli.group()
def create():
    """ create resource """
    pass


@create.command('dataset')
@click.argument('name')
@click.argument('sim')
@click.argument('args', nargs=-1)
def create_dataset(name, sim, args):
    dataset_config = parse_args(args)
    create_generated_dataset(name, sim, dataset_config)


@create.command('job')
@click.argument('name')
@click.argument('operation')
@click.option('filters', '-f', multiple=True)
@click.option('configfile', '-configfile')
@click.option('sweepfile', '-sweepfile')
@click.argument('args', nargs=-1)
def create_job(name, operation, filters, configfile, sweepfile, args):
    datasets_list = []
    for dfilter in filters:
        datasets_list.extend(filter_dataset(dfilter))

    if sweepfile:
        sweep_config = read_json(sweepfile)
        bindings = sweep_config['gin_bindings']
        for c, random_binding in enumerate([dict(zip(bindings, v)) for v in product(*bindings.values())]):
            job_name = f'{name} {c}'
            job_config = deepcopy(sweep_config)
            job_config['gin_bindings'] = random_binding
            create_new_job(job_name, operation, job_config, datasets_list)
        return

    if configfile:
        job_config = read_json(configfile)
    else:
        job_config = parse_args(args)
    create_new_job(name, operation, job_config, datasets_list)


@create.command('sweep')
@click.argument('name')
@click.argument('sim')
@click.argument('number')
@click.argument('args', nargs=-1)
def create_sweep(name, sim, number, args):
    dataset_config = parse_args(args)
    for i in range(int(number)):
        dataset_name = f'{name} seed{i}'
        dataset_config['seed'] = i
        create_generated_dataset(dataset_name, sim, dataset_config)



"""
    Base class for new annotation formats. Any new annotation format must be
    able to take a saver object and output annotations.
"""
import logging
from pathlib import Path
from typing import Dict, Union

import zpy

log = logging.getLogger(__name__)


class Output:
    """ Outputs a Saver object to various annotation file formats. """

    def __init__(self,
                 saver: zpy.saver.Saver = None,
                 annotation_filename: Union[Path, str] = 'annotations.txt',
                 ):
        """ Create an Output object from a Saver object.

        Args:
            saver (zpy.saver.Saver, optional): Saver contains annotation and category information.
            annotation_filename (Union[Path, str], optional): Default name for annotation files.
        """
        if saver is not None:
            self.saver = saver
            # Try and deduce the annotation path from the saver object
            if self.saver.annotation_path is not None:
                self.annotation_path = self.saver.annotation_path
            elif self.saver.output_dir is not None:
                self.annotation_path = self.saver.output_dir / annotation_filename
            else:
                self.annotation_path = None
                log.warning(
                    'No annotation path could be deduced from Saver object.')

    def output_annotations(self,
                            annotation_path: Union[Path, str] = None,
                            ) -> Path:
        """ Output annotations to file.

        Args:
            annotation_path (Union[Path, str], optional): Output path for annotation file.

        Returns:
            Path: Path to annotation file.
        """
        if annotation_path is None:
            annotation_path = self.annotation_path
        log.info(f'Outputting annotation file to {annotation_path}')
        annotation_path = zpy.files.verify_path(annotation_path)
        return annotation_path


"""
    COCO (Common Objects in Context) dataset format.
"""
import copy
import logging
from pathlib import Path
from typing import List, Union

import gin
import zpy

log = logging.getLogger(__name__)


class COCOParseError(Exception):
    """ Invalid COCO Annotation found when parsing data contents. """
    pass


@gin.configurable
class OutputCOCO(zpy.output.Output):
    """ Output class for COCO style annotations.

    https://cocodataset.org/#home

    """

    ANNOTATION_FILENAME = Path('coco_annotations.json')

    def __init__(self, *args, **kwargs) -> Path:
        super().__init__(*args, annotation_filename=self.ANNOTATION_FILENAME, **kwargs)

    @gin.configurable
    def output_annotations(self,
                           annotation_path: Union[Path, str] = None,
                           splitseg: bool = False,
                           ) -> Path:
        """ Output COCO annotations to file.

        Args:
            annotation_path (Union[Path, str], optional): Output path for annotation file.
            splitseg (bool, optional): Optionally output split-segmentation annotations. Defaults to False.

        Returns:
            Path: Path to annotation file.
        """
        annotation_path = super().output_annotations(annotation_path=annotation_path)
        coco_dict = {
            'info': self.coco_info(),
            'licenses': self.coco_license(),
            'categories': self.coco_categories(),
            'images': self.coco_images(),
            'annotations': self.coco_annotations(),
        }
        # Write out annotations to file
        zpy.files.write_json(annotation_path, coco_dict)
        parse_coco_annotations(annotation_path, data_dir=self.saver.output_dir)
        # Output split-segmentation annotations
        if splitseg:
            log.info('Outputting COCO annotations with multi-part' +
                     'segmentation split into seperate annotations')
            coco_dict['annotations'] = self.coco_split_segmentation_annotations()
            annotation_path = zpy.files.make_underscore_path(
                annotation_path, 'splitseg')
            # Write out annotations to file
            zpy.files.write_json(annotation_path, coco_dict)
            parse_coco_annotations(
                annotation_path, data_dir=self.saver.output_dir)
        return annotation_path

    @gin.configurable
    def coco_info(self,
                  keys_to_add: List[str] = None
                  ):
        """ coco info """
        coco_info = {
            'description': self.saver.metadata['description'],
            'url': self.saver.metadata['url'],
            'version': self.saver.metadata['date_created'],
            'year': self.saver.metadata['year'],
            'contributor': self.saver.metadata['contributor'],
            'date_created': self.saver.metadata['date_created'],
            'save_path': self.saver.metadata['save_path'],
        }
        # Add any extra keys.
        if keys_to_add is not None:
            for key in keys_to_add:
                value = self.saver.metadata.get(key, None)
                if value is not None:
                    coco_info[key] = value
        return coco_info

    def coco_license(self):
        """ coco license """
        return {
            "url": "http://zumolabs.ai/image_license/",
            "id": 0,
            "name": "Zumo Labs Image License"
        }

    @gin.configurable
    def coco_categories(self,
                        keys_to_add: List[str] = [
                            'keypoints',
                            'skeleton',
                            'supercategory',
                            'subcategories',
                        ]
                        ):
        """ coco categories """
        coco_categories = []
        for category in self.saver.categories.values():
            coco_category = {
                'id': category['id'],
                'name': category['name'],
            }
            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = category.get(key, None)
                    if value is not None:
                        coco_category[key] = value
            coco_categories.append(coco_category)
        return coco_categories

    @gin.configurable
    def coco_images(self,
                    only_default_images: bool = True,
                    keys_to_add: List[str] = None,
                    ):
        """ coco images """
        _coco_img_id_to_idx = {}  # Image id does not correspond to index in image list
        coco_images = []
        for image in self.saver.images.values():
            if only_default_images and not image['style'] == 'default':
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_img = {
                'license': 0,
                'id': image['id'],
                'file_name': image['name'],
                'coco_url': image['name'],
                'width': image['width'],
                'height': image['height'],
                'date_captured': self.saver.metadata['date_created'],
                'flickr_url': '.',
            }
            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = image.get(key, None)
                    if value is not None:
                        coco_img[key] = value
            coco_images.append(coco_img)
        return coco_images

    @gin.configurable
    def coco_annotations(self,
                         keys_to_add: List[str] = [
                             'bbox', 'area', 'segmentation'],
                         clipped: bool = True,
                         only_default_images: bool = True,
                         ):
        """ coco annotations """
        coco_annotations = []
        for annotation in self.saver.annotations:
            if only_default_images and \
                    not self.saver.images[annotation['image_id']]['style'] == 'default':
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_ann = {
                'category_id': annotation['category_id'],
                'image_id': annotation['image_id'],
                'id': annotation['id'],
                'iscrowd': False,
            }
            if clipped:
                height = self.saver.images[annotation['image_id']]['height']
                width = self.saver.images[annotation['image_id']]['width']

            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = annotation.get(key, None)
                    if value is not None:
                        if key == 'segmentation':
                            coco_ann['segmentation'] = self.saver.clip_coordinate_list(
                                width=width,
                                height=height,
                                annotation=annotation['segmentation']) if clipped \
                                else annotation['segmentation']
                        elif key == 'segmentation_rle':
                            coco_ann['segmentation_rle'] = annotation['segmentation_rle']
                        elif key == 'segmentation_float':
                            coco_ann['segmentation_float'] = self.saver.clip_coordinate_list(
                                normalized=True,
                                annotation=annotation['segmentation_float']) if clipped \
                                else annotation['segmentation_float']
                        elif key == 'bbox_float':
                            coco_ann['bbox_float'] = self.saver.clip_bbox(
                                normalized=True,
                                bbox=annotation['bbox_float']) if clipped \
                                else annotation['bbox_float']
                        elif key == 'bbox':
                            coco_ann['bbox'] = self.saver.clip_bbox(
                                width=width,
                                height=height,
                                bbox=annotation['bbox']) if clipped \
                                else annotation['bbox']
                        elif key == 'bboxes_float':
                            coco_ann['bboxes_float'] = [self.saver.clip_bbox(
                                normalized=True,
                                bbox=bbox) for
                                bbox in annotation['bboxes_float']] if clipped \
                                else annotation['bboxes_float']
                        elif key == 'bboxes':
                            coco_ann['bboxes'] = [self.saver.clip_bbox(
                                height=height,
                                width=width,
                                bbox=bbox) for
                                bbox in annotation['bboxes']] if clipped \
                                else annotation['bboxes']
                        else:
                            coco_ann[key] = value
                    try:
                        if key == 'area':
                            coco_ann['area'] = annotation['bbox'][2] * \
                                annotation['bbox'][3]
                        elif key == 'areas':
                            coco_ann['areas'] = [bbox[2] * bbox[3]
                                                 for bbox in annotation['bboxes']]
                    except:
                        pass
            # HACK: Require bbox for an annotation
            if coco_ann.get('bbox', None) is not None:
                coco_annotations.append(coco_ann)

        return coco_annotations

    @gin.configurable
    def coco_split_segmentation_annotations(self,
                                            keys_to_add: List[str] = [
                                                'bbox', 'area', 'segmentation'],
                                            clipped: bool = True,
                                            only_default_images: bool = True,
                                            ):
        """ coco annotations one per segmentation """
        coco_annotations = []
        # Annotation id will be re-mapped
        annotation_id = 0
        for annotation in self.saver.annotations:
            if only_default_images and \
                    not self.saver.images[annotation['image_id']]['style'] == 'default':
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_ann = {
                'category_id': annotation['category_id'],
                'image_id': annotation['image_id'],
                'iscrowd': False,
            }
            if clipped:
                height = self.saver.images[annotation['image_id']]['height']
                width = self.saver.images[annotation['image_id']]['width']

            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = annotation.get(key, None)
                    if value is not None:
                        coco_ann[key] = value

            # Annotations can be composed of multiple annotation components
            if annotation.get('segmentation') is not None:
                num_components = len(annotation['segmentation'])
            else:
                log.warning(
                    'Skipping annotation: split segmentation requires segmentaiton field.')
                continue

            # TODO: This can prolly be cleaned up?
            for i in range(num_components):
                _coco_ann = copy.deepcopy(coco_ann)
                try:
                    _coco_ann['segmentation'] = [self.saver.clip_coordinate_list(
                        height=height,
                        width=width,
                        annotation=annotation['segmentation'][i])] if clipped \
                        else [annotation['segmentation'][i]]
                except:
                    pass
                try:
                    _coco_ann['segmentation_rle'] = [
                        annotation['segmentation_rle'][i]]
                except:
                    pass
                try:
                    _coco_ann['segmentation_float'] = [self.saver.clip_coordinate_list(
                        normalized=True,
                        annotation=annotation['segmentation_float'][i])] if clipped \
                        else [annotation['segmentation_float'][i]]
                except:
                    pass
                try:
                    _coco_ann['bbox_float'] = self.saver.clip_bbox(
                        normalized=True,
                        bbox=annotation['bboxes_float'][i]) if clipped \
                        else annotation['bboxes_float'][i]
                except:
                    pass
                try:
                    _coco_ann['bbox'] = self.saver.clip_bbox(
                        width=width,
                        height=height,
                        bbox=annotation['bboxes'][i]) if clipped \
                        else annotation['bboxes'][i]
                except:
                    pass
                try:
                    _coco_ann['area'] = annotation['areas'][i]
                except:
                    pass

                # HACK: Require bbox for an annotation
                if _coco_ann.get('bbox', None) is not None:
                    _coco_ann['id'] = annotation_id
                    annotation_id += 1
                    coco_annotations.append(_coco_ann)

        return coco_annotations


@gin.configurable
def parse_coco_annotations(
    annotation_file: Union[Path, str],
    data_dir: Union[Path, str] = None,
    output_saver: bool = False,
    # Specify which keys to add to ImageSaver
    image_keys_to_add: List[str] = None,
) -> zpy.saver_image.ImageSaver:
    """ Parse COCO annotations, optionally output a ImageSaver object.

    Args:
        annotation_file (Union[Path, str]): Path to annotation file.
        data_dir (Union[Path, str], optional): Directory containing data (images, video, etc).
        output_saver (bool, optional): Whether to return a Saver object or not. Defaults to False.
        image_keys_to_add (List[str], optional): Image dictionary keys to include when parsing COCO dict.

    Raises:
        COCOParseError: Various checks on annotations, categories, images

    Returns:
        zpy.saver_image.ImageSaver: Saver object for Image datasets.
    """
    log.info(f'Parsing COCO annotations at {annotation_file}...')
    # Check annotation file path
    annotation_file = zpy.files.verify_path(annotation_file)
    if data_dir is not None:
        data_dir = zpy.files.verify_path(data_dir, check_dir=True)
    else:
        # If no data_dir, assume annotation file is in the root folder.
        data_dir = annotation_file.parent

    # Check that categories, images, and annotations are not blank
    coco_annotations = zpy.files.read_json(annotation_file)
    images = coco_annotations["images"]
    if len(images) == 0:
        raise COCOParseError(f'no images found in {annotation_file}')
    categories = coco_annotations["categories"]
    if len(categories) == 0:
        raise COCOParseError(
            f'no categories found in {annotation_file}')
    annotations = coco_annotations["annotations"]
    if len(annotations) == 0:
        raise COCOParseError(
            f'no annotations found in {annotation_file}')
    log.info(
        f'images:{len(images)} categories:{len(categories)} annotations:{len(annotations)}')

    # Optionally output a saver object.
    if output_saver:
        saver = zpy.saver_image.ImageSaver(output_dir=data_dir,
                                           annotation_path=annotation_file,
                                           description=coco_annotations['info']['description'],
                                           clean_dir=False,
                                           )

    # Check Images
    log.info('Parsing images...')
    img_ids = []
    for img in images:
        # Image ID
        image_id = img['id']
        if not isinstance(image_id, int):
            raise COCOParseError(f'image id {image_id} must be int.')
        if image_id in img_ids:
            raise COCOParseError(f'image id {image_id} already used.')
        img_ids.append(image_id)
        if image_id < 0:
            raise COCOParseError(f'invalid image id {image_id}')
        # Height and Width
        height, width = img['height'], img['width']
        if not isinstance(height, int):
            raise COCOParseError(f'height {height} must be int.')
        if not isinstance(width, int):
            raise COCOParseError(f'width {width} must be int.')
        if height <= 0 or width <= 0:
            raise COCOParseError(
                f'width and height h:{height} w:{width} must be > 0')
        # Image Name
        filename = img['file_name']
        if not isinstance(filename, str):
            raise COCOParseError(f'filename {filename} must be str.')
        image_path = data_dir / filename
        if not image_path.exists():
            raise COCOParseError(
                f'image path {image_path} does not exist')
        # COCO Path
        coco_url = img.get('coco_url', None)
        if coco_url is None:
            coco_url = filename
        coco_url = Path(coco_url)
        coco_path = data_dir / coco_url
        if not coco_path.exists():
            raise COCOParseError(f'coco url {coco_path} does not exist')
        # Save each image to ImageSaver object
        if output_saver:
            saver.images[image_id] = {
                'id': image_id,
                'name': filename,
                'output_path': str(coco_url),
                'height': height,
                'width': width,
                'style': 'default',
            }
            # Add any extra keys.
            if image_keys_to_add is not None:
                for key in image_keys_to_add:
                    value = img.get(key, None)
                    if value is not None:
                        saver.images[image_id][key] = value

    # Check Categories
    log.info('Parsing categories...')
    cat_ids = []
    cat_names = []
    for category in categories:
        name, category_id = category['name'], category['id']
        log.info(f'name:{name} id:{category_id}')
        # Category Name
        category_name = category['name']
        if not isinstance(category_name, str):
            raise COCOParseError(
                f'category_name {category_name} must be str.')
        if category_name in cat_names:
            raise COCOParseError(
                f'category_name {category_name} already used')
        cat_names.append(category_name)
        # Category ID
        category_id = category['id']
        if not isinstance(image_id, int):
            raise COCOParseError(
                f'category_id {category_id} must be int.')
        if category_id in cat_ids:
            raise COCOParseError(
                f'category id {category_id} already used')
        cat_ids.append(category_id)
        # Supercategories
        if category.get('supercategory', None) is not None:
            pass
        if category.get('supercategories', None) is not None:
            pass
        # Subcategories
        if category.get('subcategory', None) is not None:
            pass
        if category.get('subcategories', None) is not None:
            pass
        # Keypoints
        if category.get("keypoints", None) is not None:
            keypoints = category['keypoints']
            log.info(f"{len(keypoints)} keypoints:{keypoints}")
            if category.get('skeleton', None) is None:
                raise COCOParseError(
                    f'skeleton must be present with {keypoints}')
        # Save each category to ImageSaver object
        if output_saver:
            _category = saver.categories.get(category_id, None)
            if _category is None:
                saver.categories[category_id] = {}
            for key, value in category.items():
                saver.categories[category_id][key] = value

    # Check Annotations
    log.info('Parsing annotations...')
    ann_ids = []
    for annotation in annotations:
        # IDs
        image_id, category_id, annotation_id = annotation[
            'image_id'], annotation['category_id'], annotation['id']
        if image_id not in img_ids:
            raise COCOParseError(
                f'annotation img:{image_id} not in {img_ids}')
        if category_id not in cat_ids:
            raise COCOParseError(
                f'annotation cat:{category_id} not in {cat_ids}')
        if annotation_id in ann_ids:
            raise COCOParseError(
                f'annotation id:{annotation_id} already used')
        ann_ids.append(annotation_id)

        # Bounding Boxes
        bbox = annotation.get('bbox', None)
        if bbox is not None:
            pass

        # Keypoints
        keypoints = annotation.get('num_keypoints', None)
        if keypoints is not None:
            if "keypoints_xyv" in annotation:
                if len(annotation["keypoints_xyv"]) != int(annotation["num_keypoints"])*3:
                    raise COCOParseError(
                        'keypoints_xyv not correct size {len(keypoints)}')
            if "keypoints_xyz" in annotation:
                if len(annotation["keypoints_xyz"]) != int(annotation["num_keypoints"])*3:
                    raise COCOParseError(
                        'keypoints_xyz not correct size {len(keypoints)}')

        # Save each annotation to ImageSaver object
        if output_saver:
            saver.annotations.append(annotation)

    if output_saver:
        return saver
    else:
        return None



"""
    Logging utilities.
"""
from pathlib import Path
from typing import List, Union
import logging
import shutil

log = logging.getLogger(__name__)


def set_log_levels(
    level: str = None,
    modules: List[str] = [
        'zpy',
        'zpy_addon',
        'bpy.zpy_addon'
        'neuralyzer',
        'bender',
    ],
) -> None:
    """ Set logger levels for all zpy modules.

    Args:
        level (str, optional): log level in [info, debug, warning]. Defaults to logging.Info.
        modules (List[str], optional): Modules to set logging for. Defaults to [ 'zpy', 'zpy_addon', 'bpy.zpy_addon' 'neuralyzer', ].
    """
    if level is None:
        log_level = logging.INFO
    elif level == 'info':
        log_level = logging.INFO
    elif level == 'debug':
        log_level = logging.DEBUG
    elif level == 'warning':
        log_level = logging.WARNING
    else:
        log.warning(f'Invalid log level {level}')
        return
    log.warning(f'Setting log level to {log_level} ({level})')
    logging.basicConfig(
        level=log_level,
        format='%(asctime)s: %(levelname)s %(filename)s] %(message)s')
    for logger_name in modules:
        try:
            logging.getLogger(logger_name).setLevel(log_level)
        except:
            pass


def linebreaker_log(
    message: str,
    line_length: int = 80,
):
    """ Good looking line-breaker log message.

    Args:
        message (str): Message to put in the log.
        line_length (int, optional): Length of line breakers ----. Defaults to 80.
    """
    # Clip the message
    message = message[:line_length]
    whitespace = ' ' * int((line_length - len(message)) / 2)
    # La piece de resistance
    log.info('-'*line_length)
    log.info(f'{whitespace}{message.upper()}{whitespace}')
    log.info('-'*line_length)


def setup_file_handlers(
    log_dir: Union[str, Path] = '/tmp',
    error_log: bool = True,
    debug_log: bool = True,
) -> None:
    """ Output log files for requests

    Args:
        error_log: output error.log
        debug_log: output debug.log
        log_dir: directory to output log files
    """
    root = logging.getLogger()

    info_fh = logging.FileHandler(f"{log_dir}/info.log", mode="w")
    info_fh.setLevel(logging.INFO)
    root.addHandler(info_fh)

    if error_log:
        error_fh = logging.FileHandler(f"{log_dir}/error.log", mode="w")
        error_fh.setLevel(logging.ERROR)
        root.addHandler(error_fh)

    if debug_log:
        debug_fh = logging.FileHandler(f"{log_dir}/debug.log", mode="w")
        debug_fh.setLevel(logging.DEBUG)
        root.addHandler(debug_fh)


def  save_log_files(
    output_dir: Union[str, Path],
    log_dir: Union[str, Path] = '/tmp',
) -> None:
    """ Save log files to output directory

    Args:
        output_dir: directory to save log files
        log_dir: directory where logs exist
    """
    for log in ['info.log', 'debug.log', 'error.log']:
        src = Path(log_dir) / log
        dst = Path(output_dir) / log
        if src.exists() and src != dst:
            shutil.copy(src, dst)



"""
    HDRI utilities. You can find lots of free HDRIs at https://hdrihaven.com/
"""
import logging
import math
import os
import random
from pathlib import Path
from typing import Dict, List, Tuple, Union

import bpy
import mathutils

import gin
import zpy

log = logging.getLogger(__name__)


@gin.configurable
def load_hdri(
    path: Union[Path, str],
    scale: Tuple[float] = (1.0, 1.0, 1.0),
    random_z_rot: bool = True,
) -> None:
    """ Load an HDRI from path.

    Args:
        path (Union[Path, str]): Path to the HDRI.
        scale (Tuple[float], optional): Scale in (x, y, z). Defaults to (1.0, 1.0, 1.0).
        random_z_rot (bool, optional): Randomly rotate HDRI around Z axis. Defaults to True.
    """
    scene = zpy.blender.verify_blender_scene()
    scene.world.use_nodes = True
    tree = scene.world.node_tree
    out_node = zpy.nodes.get_or_make(
        'World Output', 'ShaderNodeOutputWorld', tree, pos=(0, 0))
    bg_node = zpy.nodes.get_or_make(
        'Background', 'ShaderNodeBackground', tree, pos=(-150, 0))
    env_node = zpy.nodes.get_or_make(
        'Environment Texture', 'ShaderNodeTexEnvironment', tree, pos=(-400, 0))
    log.info(f'Loading HDRI at {path}')
    path = zpy.files.verify_path(path, make=False)
    env_node.image = bpy.data.images.load(str(path))
    env_node.texture_mapping.scale = mathutils.Vector(scale)
    world_rot_node = zpy.nodes.get_or_make(
        'World Rotation', 'ShaderNodeVectorRotate', tree, pos=(-550, 0))
    world_rot_node.rotation_type = 'Z_AXIS'
    if random_z_rot:
        world_rotation = random.uniform(0, math.pi)
        log.debug(f'Rotating HDRI randomly along Z axis to {world_rotation}')
        world_rot_node.inputs['Angle'].default_value = world_rotation
    texcoord_node = zpy.nodes.get_or_make(
        'Texture Coordinate', 'ShaderNodeTexCoord', tree, pos=(-730, 0))
    # Link all the nodes together
    tree.links.new(out_node.inputs[0], bg_node.outputs[0])
    tree.links.new(bg_node.inputs[0], env_node.outputs[0])
    tree.links.new(env_node.inputs[0], world_rot_node.outputs[0])
    tree.links.new(world_rot_node.inputs[0], texcoord_node.outputs[0])


@gin.configurable
def random_hdri(
    hdri_dir: Union[Path, str] = 'lib/hdris/1k',
    relative_to_assets_dir : bool = True,
    apply_to_scene: bool = True,
) -> Path:
    """ Generate a random HDRI from an asset path.

    Args:
        hdri_dir (Union[Path, str], optional): Path to directory with HDRIs.
        relative_to_assets_dir (bool, optional): Path is relative to the $ASSETS directory. Defaults to False.
        apply_to_scene (bool, optional): Load the HDRI into the active scene. Defaults to True.

    Returns:
        Path: Path to the random HDRI.
    """
    if relative_to_assets_dir:
        hdri_dir = zpy.blender.get_asset_lib_path().joinpath(hdri_dir)
    hdri_dir = zpy.files.verify_path(hdri_dir, check_dir=True)
    # Create list of HDRIs in directory
    hdri_paths = []
    for _path in hdri_dir.iterdir():
        if _path.is_file() and _path.suffix in ['.hdri', '.hdr']:
            hdri_paths.append(_path)
    hdri_path = random.choice(hdri_paths)
    log.info(f'Found {len(hdri_paths)} HDRIs at {hdri_dir}')
    log.info(f'Randomly picked {hdri_path.stem}')
    if apply_to_scene:
        load_hdri(hdri_path)
    return hdri_path



"""
    Utilities for Materials in Blender.
"""
import logging
import os
import copy
import random
from pathlib import Path
from typing import Tuple, Union

import bpy
import numpy as np

import gin
import zpy

log = logging.getLogger(__name__)


def verify(
    mat: Union[bpy.types.Material, str],
    check_none: bool = True,
) -> bpy.types.Material:
    """ Get a material given either its name or the object itself.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
        check_none (bool, optional): Check to make sure material exists. Defaults to True.

    Raises:
        ValueError: Material does not exist.

    Returns:
        bpy.types.Material: Material object.
    """
    if isinstance(mat, str):
        mat = bpy.data.materials.get(mat)
    if check_none and mat is None:
        raise ValueError(f'Could not find material {mat}.')
    return mat


def for_mat_in_obj(
    obj: Union[bpy.types.Object, str],
) -> bpy.types.Material:
    """ Yield materials in scene object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)

    Raises:
        ValueError: [description]

    Returns:
        bpy.types.Material: Material object.
    """
    obj = zpy.objects.verify(obj)
    if len(obj.material_slots) > 1:
        for mat in obj.material_slots:
            yield mat.material
    else:
        if obj.active_material is not None:
            return obj.active_material
        else:
            log.debug(
                f'No active material or material slots found for {obj.name}')
            return None


_SAVED_MATERIALS = {}


def save_mat_props(
    mat: Union[bpy.types.Material, str],
) -> None:
    """ Save a pose (rot and pos) to dict.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
    """
    log.info(f'Saving material properties for {mat.name}')
    _SAVED_MATERIALS[mat.name] = get_mat_props(mat)


def restore_mat_props(
    mat: Union[bpy.types.Material, str],
) -> None:
    """ Restore an object to a position.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
    """
    log.info(f'Restoring material properties for {mat.name}')
    set_mat_props(mat, _SAVED_MATERIALS[mat.name])


def restore_all_mat_props() -> None:
    """ Restore all jittered materials to original look. """
    for mat_name, mat_props in _SAVED_MATERIALS.items():
        set_mat_props(mat_name, mat_props)


def get_mat_props(
    mat: Union[bpy.types.Material, str],
) -> Tuple[float]:
    """ Get (some of the) material properties.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)

    Returns:
        Tuple[float]: Material property values (roughness, metallic, specular).
    """
    mat = verify(mat)
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF')
    if bsdf_node is None:
        log.warning(f'No BSDF node in {mat.name}')
        return (0.0, 0.0, 0.0)
    return (
        bsdf_node.inputs['Roughness'].default_value,
        bsdf_node.inputs['Metallic'].default_value,
        bsdf_node.inputs['Specular'].default_value,
    )


def set_mat_props(
    mat: Union[bpy.types.Material, str],
    prop_tuple: Tuple[float],
) -> None:
    """ Set (some of the) material properties.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
        prop_tuple (Tuple[float]): Material property values (roughness, metallic, specular).
    """
    mat = verify(mat)
    # TODO: Work backwards from Material output node instead of
    #       assuming a 'Principled BSDF' node
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF', None)
    if bsdf_node is None:
        log.warning(f'No BSDF node in {mat.name}')
        return
    bsdf_node.inputs['Roughness'].default_value = copy.copy(prop_tuple[0])
    bsdf_node.inputs['Metallic'].default_value = copy.copy(prop_tuple[1])
    bsdf_node.inputs['Specular'].default_value = copy.copy(prop_tuple[2])


@gin.configurable
def jitter(
    mat: Union[bpy.types.Material, str],
    std: float = 0.2,
    save_first_time: bool = True,
) -> None:
    """ Randomize an existing material a little.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
        std (float, optional): Standard deviation of gaussian noise over material property. Defaults to 0.2.
        save_first_time (bool, optional): Save the material props first time jitter is called and restore before jittering every subsequent time. Defaults to True.
    """
    mat = verify(mat)
    if save_first_time:
        if _SAVED_MATERIALS.get(mat.name, None) is None:
            save_mat_props(mat)
        else:
            restore_mat_props(mat)
    log.info(f'Jittering material {mat.name}')
    mat_props = get_mat_props(mat)
    jittered_mat_props = tuple(
        map(lambda p: p + random.gauss(0, std), mat_props))
    set_mat_props(mat, jittered_mat_props)


@gin.configurable
def random_texture_mat(
    texture_dir: Union[Path, str] = 'lib/textures/random_512p',
    relative_to_assets_dir: bool = True,
) -> bpy.types.Material:
    """ Generate a random material from a directory of random texture images.

    Args:
        texture_dir (Union[Path, str], optional): Path to directory with texture images.
        relative_to_assets_dir (bool, optional): Path is relative to the $ASSETS directory. Defaults to False.

    Returns:
        bpy.types.Material: The newly created material.
    """
    if relative_to_assets_dir:
        texture_dir = zpy.blender.get_asset_lib_path().joinpath(texture_dir)
    texture_dir = zpy.files.verify_path(texture_dir, check_dir=True)
    # Create list of texture images in directory
    texture_paths = []
    for _path in texture_dir.iterdir():
        if _path.is_file() and _path.suffix in ['.jpg', '.png']:
            texture_paths.append(_path)
    texture_path = random.choice(texture_paths)
    log.info(f'Found {len(texture_paths)} Textures at {texture_dir}')
    log.info(f'Randomly picked {texture_path.stem}')
    return make_mat_from_texture(texture_path, name=texture_path.stem)


@gin.configurable
def make_mat_from_texture(
    texture_path: Union[Path, str],
    name: str = None,
) -> bpy.types.Material:
    """ Makes a material from a texture image.

    Args:
        texture_path (Union[Path, str]): Path to texture image.
        name (str, optional): Name of new material.

    Returns:
        bpy.types.Material: The newly created material.
    """
    texture_path = zpy.files.verify_path(texture_path, make=False)
    if name is None:
        name = texture_path.stem
    mat = bpy.data.materials.get(name, None)
    if mat is None:
        log.debug(f'Material {name} does not exist, creating it.')
        mat = bpy.data.materials.new(name=name)
    mat.use_nodes = True
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF')
    out_node = mat.node_tree.nodes.get('Material Output')
    tex_node = mat.node_tree.nodes.new('ShaderNodeTexImage')
    tex_node.name = 'ImageTexture'
    coord_node = mat.node_tree.nodes.new('ShaderNodeTexCoord')
    bpy.ops.image.open(filepath=str(texture_path))
    tex_node.image = bpy.data.images[texture_path.name]
    tex_node.image.colorspace_settings.name = 'Filmic Log'
    mat.node_tree.links.new(tex_node.outputs[0], bsdf_node.inputs[0])
    mat.node_tree.links.new(coord_node.outputs[2], tex_node.inputs[0])
    mat.node_tree.links.new(out_node.inputs[0], bsdf_node.outputs[0])
    tex_node.image.reload()
    return mat


@gin.configurable
def make_mat_from_color(
    color: Tuple[float],
    name: str = None,
) -> bpy.types.Material:
    """ Makes a material given a color.

    Args:
        color (Tuple[float]): Color tuple (RGB).
        name (str, optional): Name of new material.

    Returns:
        bpy.types.Material: The newly created material.
    """
    if name is None:
        name = str(color)
    mat = bpy.data.materials.get(name, None)
    if mat is None:
        log.debug(f'Material {name} does not exist, creating it.')
        mat = bpy.data.materials.new(name=name)
    mat.use_nodes = True
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF')
    out_node = mat.node_tree.nodes.get('Material Output')
    mat.node_tree.nodes.remove(bsdf_node)
    bsdf_node = mat.node_tree.nodes.new('ShaderNodeBsdfDiffuse')
    bsdf_node.inputs['Color'].default_value = color + (1.,)
    mat.node_tree.links.new(out_node.inputs[0], bsdf_node.outputs[0])
    return mat


def set_mat(
    obj: Union[bpy.types.Object, str],
    mat: Union[bpy.types.Material, str],
    recursive: bool = True,
) -> None:
    """ Set the material for an object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name) with an active material.
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
        recursive (bool, optional): Recursively set material for child objects. Defaults to True.
    """
    obj = zpy.objects.verify(obj)
    mat = zpy.material.verify(mat)
    if hasattr(obj, 'active_material'):
        log.debug(f'Setting object {obj.name} material {mat.name}')
        obj.active_material = mat
    else:
        log.warning('Object does not have material property')
        return
    # Recursively change material on all children of object
    if recursive:
        for child in obj.children:
            set_mat(child, mat)


@gin.configurable
def make_aov_material_output_node(
    mat: bpy.types.Material = None,
    obj: bpy.types.Object = None,
    style: str = 'instance',
) -> None:
    """ Make AOV Output nodes in Composition Graph.

    Args:
        mat (bpy.types.Material, optional): A blender material (either it's name or the object itself).
        obj (bpy.types.Object, optional): A blender object (either it's name or the object itself).
        style (str, optional): Type of segmentation in [instance, category]. Defaults to 'instance'.

    Raises:
        ValueError: Invalid style, no object or material given.
    """
    # Make sure engine is set to Cycles
    scene = zpy.blender.verify_blender_scene()
    if not (scene.render.engine == "CYCLES"):
        log.warning(' Setting render engine to CYCLES to use AOV')
        scene.render.engine == "CYCLES"

    # TODO: Refactor this legacy "styles" code

    # Only certain styles are available
    valid_styles = ['instance', 'category']
    assert style in valid_styles, \
        f'Invalid style {style} for AOV material output node, must be in {valid_styles}.'

    # HACK: multiple material slots
    all_mats = []

    # Use material
    if mat is not None:
        all_mats = [mat]
    # Get material from object
    elif obj is not None:
        if obj.active_material is None:
            log.debug(f'No active material found for {obj.name}')
            return
        if len(obj.material_slots) > 1:
            for mat in obj.material_slots:
                all_mats.append(mat.material)
        else:
            all_mats.append(obj.active_material)
    else:
        raise ValueError('Must pass in an Object or Material')

    # HACK: multiple material slots
    for mat in all_mats:

        # Make sure material is using nodes
        if not mat.use_nodes:
            mat.use_nodes = True
        tree = mat.node_tree

        # Vertex Color Node
        vcol_node = zpy.nodes.get_or_make(
            f'{style} Vertex Color', 'ShaderNodeVertexColor', tree)
        vcol_node.layer_name = style

        # AOV Output Node
        # HACK: This type of node has a "name" property which prevents using the
        # normal zpy.nodes code due to a scope conflict with the bpy.types.Node.name property
        # See: https://docs.blender.org/api/current/bpy.types.ShaderNodeOutputAOV.html
        _name = style
        aovout_node = None
        for _node in tree.nodes:
            if _node.name == _name:
                aovout_node = _node
        if aovout_node is None:
            aovout_node = tree.nodes.new('ShaderNodeOutputAOV')
        aovout_node.name = style

        tree.links.new(vcol_node.outputs['Color'], aovout_node.inputs['Color'])
